/* sithwm - Minimalist Window Manager for X
 * see README for Copyright, license and other details. */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "sithwm.h"

Application *head_app;
Client *head_client;

static int getrand(int range)
{
   static unsigned int randnum;
   return range<=0?0:(randnum += 7912511)%range;
}

void place_client(Client *c)
{
   c->dyn.area.pos.x = getrand(c->screen->area.w-c->dyn.area.w)-opt_bw;
   c->dyn.area.pos.y = getrand(c->screen->area.h-c->dyn.area.h)-opt_bw;
}

char default_name[2] = "*";

void make_new_client(Window w, ScreenInfo*s, int flags)
{
   XWindowAttributes winattr;
   XGetWindowAttributes(dpy, w, &winattr);
   LOG_DEBUG("\tWindow: 0x%lx (%s) %dx%d+%d+%d\n", w, map_state_string(winattr.map_state),
	     winattr.width, winattr.height, winattr.x, winattr.y);
   if (winattr.override_redirect || (flags && winattr.map_state != IsViewable))
      return;
   
   Client *c = (Client*)calloc(1, sizeof(Client));

   /* Put it second on the list, since we may not want it to be active - if on other screen. */
   Client **inspt = &head_client;
   if (head_client)
      inspt = &(head_client->next);
   c->next = (*inspt);
   (*inspt) = c;
   c->dyn.window = w;
   c->screen = s;
   c->cmap = winattr.colormap;
   c->name = default_name;

   c->dyn.area.w = winattr.width;
   c->dyn.area.h = winattr.height;
   if ( flags || winattr.x<-20 || winattr.y<-20 || (winattr.x==0&&winattr.y==0) ||
	(winattr.x+winattr.width) > s->area.w || (winattr.y+winattr.height) > s->area.w
	) {
      place_client(c);
   } else {
      c->dyn.area.pos.x = (unsigned int)winattr.x;
      c->dyn.area.pos.y = (unsigned int)winattr.y;
   }
   toworldpos(c, c->dyn.area.pos.x, c->dyn.area.pos.y, &c->dyn.area.pos);
   LOG_DEBUG("\twindow started as %dx%d +%d+%d\n", c->dyn.area.w, c->dyn.area.h, c->dyn.area.pos.x, c->dyn.area.pos.y);

#ifdef DEBUG
   {
      Client *p;
      int i = 0;
      for (p = head_client; p; p = p->next)
	 i++;
      LOG_DEBUG("make_new_client() : new window %dx%d+%d+%d, wincount=%d\n", c->dyn.area.w, c->dyn.area.h, c->dyn.area.pos.x, c->dyn.area.pos.y, i);
   }
#endif

   XSelectInput(dpy, c->dyn.window, ColormapChangeMask | EnterWindowMask | PropertyChangeMask | FocusChangeMask);
#ifdef DEBUG
   XSync(dpy, 0);
#endif
   func_grab(c->dyn.window, 1);
   /* Read instance/class information for client and check against list
    * built with -app options */
   XClassHint *xclass = XAllocClassHint();
   if (xclass) {
      Application *a = head_app;
      XGetClassHint(dpy, w, xclass);
      LOG_DEBUG("\tClass %s %s\n", xclass->res_name?xclass->res_name:"(nollinolli)", xclass->res_class?xclass->res_class:"(nollinolli)");
      while (a) {
	 if ((!a->res_name || (xclass->res_name && !strcmp(xclass->res_name, a->res_name)))
	     && (!a->res_class || (xclass->res_class && !strcmp(xclass->res_class, a->res_class)))) {
	    if (a->geometry_mask & WidthValue)
	       c->dyn.area.w = a->area.w;
	    if (a->geometry_mask & HeightValue)
	       c->dyn.area.h = a->area.h ;
	    if (a->geometry_mask & XValue) {
	       if (a->geometry_mask & XNegative)
		  c->dyn.area.pos.x = a->area.pos.x + s->area.w-c->dyn.area.w;
	       else
		  c->dyn.area.pos.x = a->area.pos.x;
	    }
	    if (a->geometry_mask & YValue) {
	       if (a->geometry_mask & YNegative)
		  c->dyn.area.pos.y = a->area.pos.y + s->area.h-c->dyn.area.h;
	       else
		  c->dyn.area.pos.y = a->area.pos.y;
	    }
	    c->dyn.sticky = a->sticky;
	 }
	 a = a->next;
      }
      XFree(xclass->res_class);
      XFree(xclass->res_name);
      XFree(xclass);
   }

   XSetWindowBorderWidth(dpy, c->dyn.window, opt_bw);
   select_client(c, (XACT_NEW<<FUNC_SHIFT)|FMOD_MAP);
}

void get_client_props(Client*client)
{
   if (client->name != default_name)
      XFree(client->name);
   XFetchName(dpy, client->dyn.window, &client->name);
   if (!client->name)
      client->name = default_name;
   XWMHints *hints = XGetWMHints(dpy, client->dyn.window);
   if (hints) {
/*       printf("%i %i\n", hints->icon_x, hints->icon_y); */
/*       if (hints->flags&IconMaskHint) */
/* 	 printf("MASK\n"); */
/*       if (hints->flags&XUrgencyHint) */
/* 	 printf("URG\n"); */
/*       client->icon_mask = (hints->flags&IconMaskHint)?hints->icon_mask:None; */

      client->icon = (hints->flags&IconWindowHint)?hints->icon_window:
	 ((hints->flags&IconPixmapHint)?hints->icon_pixmap:None);
      if (client->icon != None ) {
	 Window root_return;
	 int x_return, y_return;
	 unsigned int border_width_return,depth_return;
	 XGetGeometry(dpy, client->icon, &root_return,
		      &x_return, &y_return,
		      &client->icon_width, &client->icon_height,
		      &border_width_return, &depth_return);
      }

      XFree(hints);
   }
}

#ifdef DEBUG
const char *map_state_string(int map_state) {
	const char *map_states[4] = {
		"IsUnmapped",
		"IsUnviewable",
		"IsViewable",
		"Unknown"
	};
	return ((unsigned int)map_state < 3)
		? map_states[map_state]
		: map_states[3];
}
#endif

/* used all over the place.  return the client that has specified window */

Client *find_client(Window w) {
   Client *c;
   for (c = head_client; c; c = c->next)
      if (w == c->dyn.window)
	 return c;
   return NULL;
}

Client* select_window(Window w, unsigned int func)
{
   Client*c = find_client(w);
   select_client(c, func);
   return c;
}

void select_client(Client *c, unsigned int funcmask)
{
   LOG_DEBUG("select_client() : 0x%lx Actions: 0x%x\n", c?c->dyn.window:0, funcmask);
   if (!c)
      return;

   unsigned int func = funcmask >> FUNC_SHIFT;
   if (func>=XACT_ENTER)
      funcmask = funcmask|setmodifiers[func];

   LOG_DEBUG("SEL:0x%lx ", c->dyn.window);
   prt_mask(funcmask, funcmasks);
   LOG_DEBUG("\n");

   ScreenInfo*screen=c->screen;

   if (funcmask & FMOD_MAXIMISE) {
      toviewpos(c, &c->dyn.area.pos);
      if (c->dyn.area.w > (screen->area.w-30) ) {
	 c->dyn.area.w = screen->area.w>>1;
	 c->dyn.area.h = screen->area.h;
      } else  if (c->dyn.area.h > (screen->area.h-30) ) {
	 c->dyn.area.w = screen->area.w>>1;
	 c->dyn.area.h = screen->area.h>>1;
      } else {
	 c->dyn.area.w = screen->area.w;
	 c->dyn.area.h = screen->area.h;
      }
      place_client(c);
      toworldpos(c, c->dyn.area.pos.x, c->dyn.area.pos.y, &c->dyn.area.pos);
   }

   if (funcmask & FMOD_FIX) {
      struct pos r;
      toviewpos(c, &r);
      c->dyn.sticky ^= '!';
      toworldpos(c, r.x, r.y, &c->dyn.area.pos);
   }

   if (funcmask&FMOD_SET_VIEW)
   {
      int i;
      /* try first the desktop we've got, then the last screen pos for
	 this client, then the clients position - wich should match
	 for sure. */
      for (i=0;
	   i<2;
	   i++) {
	 LOG_DEBUG("SET_VIEW: %i trying %i,%i in %i,%i\n", i,
		   c->dyn.area.pos.x, c->dyn.area.pos.y,
		   screen->area.pos.x, screen->area.pos.y );
	 if ( intersect(&screen->area, &c->dyn.area))
	    break;
	 screen->area.pos = (!i)?c->last_screen:c->dyn.area.pos;
      }
   }

   if (funcmask&FMOD_MAP)
      XMapWindow(dpy, c->dyn.window);
   if (funcmask&(FMOD_RAISE|FMOD_MAXIMISE)) {
      struct pos r;
      toviewpos(c, &r);
      XMoveResizeWindow(dpy, c->dyn.window, r.x, r.y, c->dyn.area.w, c->dyn.area.h);
      XRaiseWindow(dpy, c->dyn.window);
      if (screen->infow_func>screen->do_infow_func)
	 screen->do_infow_func = screen->infow_func;
      LOG_DEBUG("raise() : 0x%lx\n", c?c->dyn.window:0);
   }
   if (screen == current_screen) {
      if ((funcmask&FMOD_FOCUS) && c != head_client) {
	 Client *ccc;
	 for (ccc = head_client; ccc; ccc = ccc->next) {
	    if (ccc->next == c) {
	       ccc->next = c->next;
	       c->next = head_client;
	       head_client = c;
	       break;
	    }
	 }
      }
      if (funcmask&FMOD_SETMOUSE)
	 XWarpPointer(dpy, None, c->dyn.window, 0, 0, 0, 0,
		      (c->last_cursor.x>0&&c->last_cursor.x<c->dyn.area.w)?c->last_cursor.x:(c->dyn.area.w>>1),
		      (c->last_cursor.y>0&&c->last_cursor.y<c->dyn.area.h)?c->last_cursor.y:(c->dyn.area.h>>1));
      XInstallColormap(dpy, c->cmap);
   }
   get_client_props(c);
}
